import numpy as np
import numpy.ma as ma
from numpy.ma.core import MaskedArray


def pad_at_the_end(signal: np.ndarray, pad_width: int) -> MaskedArray:

    assert pad_width >= 0, f"pad_width (={pad_width}) must be positive."

    if signal.ndim == 1:
        (n_samples,) = signal.shape
        n_dims = 1
    else:
        n_samples, n_dims = signal.shape

    return ma.masked_array(
        data=np.pad(
            signal.reshape(n_samples, n_dims).astype(np.float),
            pad_width=((0, pad_width), (0, 0)),
            mode="constant",
            constant_values=(np.nan,),
        ),
        mask=np.tile([False] * n_samples + [True] * pad_width, (n_dims, 1)).T,
    )


def get_signal_shape(signal_padded: MaskedArray) -> (int, int):
    err_msg = "Wrong dimensions: {signal_padded.shape}. Expected: (n_samples, n_dims)."
    assert signal_padded.ndim == 2, err_msg

    padded_size, n_dims = signal_padded.shape
    n_samples = padded_size - signal_padded.mask[:, 0].sum()
    return (n_samples, n_dims)



from pathlib import Path

import numpy as np
import numpy.ma as ma
from numpy.ma.core import MaskedArray
from scipy.io.arff import loadarff

from sklearn.utils import Bunch
from loadmydata.padding import pad_at_the_end
from loadmydata.utils import (
    download_from_remote_uea_ucr,
    get_local_data_path,
    get_uea_ucr_download_link,
)


def load_Xy_from_arff(data_path: Path) -> (MaskedArray, np.ndarray):
    """Load (X, y) from a .arff file.

    The shape of X is (n_series, n_samples, n_dims). The shape of y is
    (n_series,).
    """

    # load from downloaded (or cached) files
    data, meta = loadarff(data_path)
    names = meta.names()
    target = names[-1]
    # are the data multivariate or univariate?
    is_multivariate = len(names) == 2
    is_univariate = not is_multivariate
    # load y, the target variable
    y = data[target].view().astype(str)
    # load X, the attributes
    if is_univariate:
        keep_col = [col for col in names if col != target]
        X_raw = data[keep_col].view()
    elif is_multivariate:
        X_raw = data[names[0]].view()

    X = list()
    for signal in X_raw:
        if is_multivariate:
            X.append(np.array(signal.tolist()).T)
        elif is_univariate:
            X.append(np.array(signal.tolist()).T.reshape(-1, 1))

    max_size = max(signal.shape[0] for signal in X)

    X = ma.stack(
        [pad_at_the_end(signal, max_size - signal.shape[0]) for signal in X]
    )

    return X, y


def load_uea_ucr_data(name: str) -> Bunch:
    """Return data for the given data set.

    The data are contained in a `DataSet` instance.
    Time series are available in the `.X_train` and `.X_test` atttributes, which
    are pandas dataframe where each **row** contains a distinct time series.
    The labels of each time series are in the `.y_train` and `.y_test`
    atttributes.

    Args:
        name (str): data set's name, e.g. `ArrowHead` (case-sensitive).

    Returns:
        [sklearn.util.Bunch]: (dict-like) X_train, X_test, y_train, y_test, url
            and description of the data set.
    """

    # download data
    download_from_remote_uea_ucr(name)
    # get data path
    data_path = get_local_data_path(name)
    data_path_train = data_path / Path(f"{name}_TRAIN.arff")
    data_path_test = data_path / Path(f"{name}_TEST.arff")
    data_path_description = data_path / Path(f"{name}.txt")
    # load X, y for train and test
    X_train, y_train = load_Xy_from_arff(data_path_train)
    X_test, y_test = load_Xy_from_arff(data_path_test)
    # load description
    with (open(data_path_description, encoding="ISO-8859-1")) as f:
        description = f.read()

    return Bunch(
        X_train=X_train,
        y_train=y_train,
        X_test=X_test,
        y_test=y_test,
        description=description,
        url=(get_uea_ucr_download_link() / (name + ".zip")),
        location=data_path.absolute().resolve(),
    )



import os
import shutil
from pathlib import Path
from urllib.request import urlretrieve
from zipfile import ZipFile

import requests
from tqdm import tqdm
from yarl import URL

from loadmydata.config import CONFIG


def get_cache_home() -> Path:
    """Return the path of the cached data directory.

    The data dir is read from the `CONFIG` variable and is created if it
    does not exists.
    """
    cache_home = CONFIG["cache_home"]
    if not cache_home.exists():
        cache_home.mkdir()
    return cache_home


def clear_data_home() -> None:
    """Delete the content of the data cache."""
    cache_home = CONFIG["cache_home"]
    if cache_home.exists():
        shutil.rmtree(cache_home)


def get_local_data_path(name: str):
    """Return the path to the local data folder.

    Args:
        name (str): data set's name, e.g. `ArrowHead` (case-sensitive).
    """
    return get_cache_home() / name


def get_uea_ucr_download_link() -> URL:
    """Return the download link to the UEA/UCR repository.

    The download link is read from `config.ini`.
    """
    return CONFIG["uea_ucr_download_link"]


def download_from_remote_uea_ucr(name: str) -> None:
    """Download and uncompress data from UEA/UCR repository.

    Args:
        name (str): data set's name, e.g. `ArrowHead` (case-sensitive).
    """
    local_cache_data = get_local_data_path(name)

    if not local_cache_data.exists():
        local_cache_data.mkdir(exist_ok=True, parents=True)

        # get archive's url
        archive_name = name + ".zip"
        local_archive_path = local_cache_data / archive_name
        remote_archive_path = get_uea_ucr_download_link() / archive_name
        response = requests.get(remote_archive_path, stream=True)
        # handle the download progress bar
        total_size_in_bytes = int(response.headers.get("content-length", 0))
        block_size = 1024  # 1 Kibibyte
        progress_bar = tqdm(
            total=total_size_in_bytes, unit="iB", unit_scale=True
        )
        # actual download
        with open(local_archive_path, "wb") as handle:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                handle.write(data)
        progress_bar.close()
        if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
            print(f"ERROR: the download of {archive_name} went wrong.")
        # uncompress the data in the data folder
        with ZipFile(local_archive_path, "r") as zf:
            zf.extractall(local_cache_data)
        # remove zip file
        os.remove(local_archive_path)
        # Check if the extracted directory contains a single sub-directory and
        # no other file.
        directory_list = [x for x in local_cache_data.iterdir() if x.is_dir()]
        non_directory_list = [
            x for x in local_cache_data.iterdir() if not x.is_dir()
        ]
        if len(directory_list) == 1 and len(non_directory_list) == 0:
            sub_dir = directory_list[0]
            for element in sub_dir.iterdir():
                shutil.move(str(element), str(sub_dir.parent))
            os.rmdir(str(sub_dir))


def is_directory_empty(dir_path: Path) -> bool:
    """Check if a directory is empty.

    Args:
        dir_path (Path): path to directory

    Returns:
        bool: True if empty, False otherwise
    """
    assert (
        dir_path.exists()
    ), f"The provided directory does not exist: '{dir_path}'."
    assert dir_path.is_dir(), f"Provide a directory path, not '{dir_path}'."
    return not any(Path(dir_path).iterdir())



import datetime as dt
from pathlib import Path

import numpy as np
import pandas as pd
import requests
from yarl import URL

from loadmydata.config import CONFIG
from loadmydata.utils import get_local_data_path

DATASET_NAME = "NYCTaxi"
DATAFILE_NAME = "nyc_taxi.csv"

DESCRIPTION = (
    "This data set contains the number of New York taxi passengers "
    "aggregated in 30 minutes buckets for the period between July 2014 and "
    "January 2015. "
    "There are five anomalies occur during the NYC marathon, Thanksgiving, "
    "Christmas, New Years day, and a snow storm.\n\n"
    "The raw data is from the NYC Taxi and Limousine Commission [1] and has "
    "been curated by [2].\n\n"
    "[1]: https://www1.nyc.gov/site/tlc/about/tlc-trip-record-data.page \n"
    "[2]: Ahmad, S., Lavin, A., Purdy, S., & Agha, Z. (2017). Unsupervised "
    "real-time anomaly detection for streaming data. Neurocomputing."
)
# Anomaly labels are taken from the followink link:
# https://github.com/numenta/NAB/blob/master/labels/combined_labels.json
ANOMALY_LABELS = [
    "2014-11-01 19:00:00",
    "2014-11-27 15:30:00",
    "2014-12-25 15:00:00",
    "2015-01-01 01:00:00",
    "2015-01-27 00:00:00",
]


def read_timestamps_str(timestamp_str: str) -> dt.datetime:
    return dt.datetime.strptime(timestamp_str, "%Y-%m-%d %H:%M:%S")


def get_nyc_taxi_download_link() -> URL:
    """Return the download link to the UEA/UCR repository.

    The download link is read from `config.ini`.
    """
    return CONFIG["nyc_taxi_download_link"] / DATAFILE_NAME


def load_nyc_taxi_dataset() -> (pd.DataFrame, np.ndarray, str):
    """Load (X, y) from a .csv file.

    The shape of X is (n_samples, 2): [timestamp, value]. The shape of y
    is (n_anomalies,).
    """
    local_cache_data = get_local_data_path(DATASET_NAME)
    local_archive_path = local_cache_data / DATAFILE_NAME
    if not local_cache_data.exists():
        local_cache_data.mkdir(exist_ok=True, parents=True)
        # get archive's url
        remote_archive_path = get_nyc_taxi_download_link()
        response = requests.get(remote_archive_path, stream=True)
        with open(local_archive_path, "w") as handle:
            print(response.text, file=handle)

    # load from downloaded (or cached) files
    X = pd.read_csv(local_archive_path, parse_dates=["timestamp"]).rename(
        {"value": "taxi_count"}, axis=1
    )
    y = np.array(
        [
            read_timestamps_str(timestamp_str)
            for timestamp_str in ANOMALY_LABELS
        ]
    )
    return X, y, DESCRIPTION



import json
import os
import shutil
from pathlib import Path
from zipfile import ZipFile

import numpy as np
import pandas as pd
import requests
from sklearn.utils import Bunch
from tqdm import tqdm
from yarl import URL

from loadmydata.config import CONFIG, HUMAN_LOCOMOTION_CODE_LIST
from loadmydata.utils import get_local_data_path, is_directory_empty

DATASET_NAME = "HumanLocomotion"
DATAFILE_NAME = "GaitData.zip"

DESCRIPTION = """This data set consists of 1020 multivariate gait signals collected with two inertial measurement units, from 230 subjects undergoing a fixed protocol:
        - standing still,
        - walking 10 m,
        - turning around,
        - walking back,
        - stopping.

In total, there are 8.5 h of gait time series. The measured population was composed of healthy subjects as well as patients with neurological or orthopedic disorders.
The start and end time stamps of more than 40,000 footsteps are available, as well as a number of contextual information about each trial. This exact data set was used in [1] to design and evaluate a step detection procedure.

The data are thoroughly described in [2].

[1] Oudre, L., Barrois-Müller, R., Moreau, T., Truong, C., Vienne-Jumeau, A., Ricard, D., Vayatis, N., & Vidal, P.-P. (2018). Template-based step detection with inertial measurement units. Sensors, 18(11).

[2] Truong, C., Barrois-Müller, R., Moreau, T., Provost, C., Vienne-Jumeau, A., Moreau, A., Vidal, P.-P., Vayatis, N., Buffat, S., Yelnik, A., Ricard, D., & Oudre, L. (2019). A data set for the study of human locomotion with inertial measurements units. Image Processing On Line (IPOL), 9.
"""


def get_code_list():
    return HUMAN_LOCOMOTION_CODE_LIST


def get_trial_filename(code: str) -> Path:
    """Returns the filename of the signal file and the metadata file.

    Code must be "{subject number}-{trial number}", e.g. "14-3".

    Args:
        code (str): code of the trial ("Patient-Trial").

    Returns:
        Path: path to the associated files.
    """
    local_cache_data = get_local_data_path(DATASET_NAME)
    filename = local_cache_data / code
    assert filename.with_suffix(
        ".csv"
    ).exists(), f"The code {code} cannot be found in the data set."
    return filename


def get_human_locomotion_download_link() -> URL:
    """Return the download link to the UEA/UCR repository.

    The download link is read from `config.ini`.
    """
    return CONFIG["human_locomotion_download_link"] / DATAFILE_NAME


def download_from_remote_human_locomotion() -> None:
    """Download and uncompress the human locomotion data set."""
    local_cache_data = get_local_data_path(DATASET_NAME)
    local_archive_path = local_cache_data / DATAFILE_NAME

    if not local_cache_data.exists():
        local_cache_data.mkdir(exist_ok=True, parents=True)

        # get archive's url
        remote_archive_path = get_human_locomotion_download_link()
        print(remote_archive_path)
        response = requests.get(remote_archive_path, stream=True)
        # handle the download progress bar
        total_size_in_bytes = int(response.headers.get("content-length", 0))
        block_size = 1024  # 1 Kibibyte
        progress_bar = tqdm(
            total=total_size_in_bytes, unit="iB", unit_scale=True
        )
        # actual download
        with open(local_archive_path, "wb") as handle:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                handle.write(data)
        progress_bar.close()
        if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
            print(f"ERROR: the download of {DATAFILE_NAME} went wrong.")
        # uncompress the data in the data folder
        with ZipFile(local_archive_path, "r") as zf:
            zf.extractall(local_cache_data)
        # remove zip file
        os.remove(local_archive_path)
        # Check if the extracted directory contains a single sub-directory and
        # no other file.
        directory_list = [x for x in local_cache_data.iterdir() if x.is_dir()]
        non_directory_list = [
            x for x in local_cache_data.iterdir() if not x.is_dir()
        ]
        if len(directory_list) == 1 and len(non_directory_list) == 0:
            sub_dir = directory_list[0]
            for element in sub_dir.iterdir():
                shutil.move(str(element), str(sub_dir.parent))
            os.rmdir(str(sub_dir))


def load_trial(code: str) -> pd.DataFrame:
    """Returns the signal of the trial.

    Args:
        code (str): code of the trial ("Patient-Trial").

    Returns:
        pd.DataFrame: Signal of the the trial, shape (n_sample, n_dimension).
    """
    fname = get_trial_filename(code)
    df = pd.read_csv(fname.with_suffix(".csv"), sep=",")
    return df


def load_metadata(code):
    """Returns the metadata of the trial.
    Parameters
    ----------
    code : str
        Code of the trial ("Patient-Trial").
    Returns
    -------
    dict
        Metadata dictionary.
    """
    fname = get_trial_filename(code)
    with open(fname.with_suffix(".json"), "r") as f:
        metadata = json.load(f)
    return metadata


def load_human_locomotion_dataset(code: str) -> Bunch:
    """Load the human locomotion data set.

    Returns:
        sklearn.utils.Bunch: (dict-like) the acceleration and angular velocity,
            the step indexes, the metadata and the description
    """
    # check if in cache, othewise download data
    download_from_remote_human_locomotion()
    # get data
    signal = load_trial(code)
    metadata = load_metadata(code)
    left_steps = np.array(metadata.pop("LeftFootActivity"))
    right_steps = np.array(metadata.pop("RightFootActivity"))

    # load from the downloaded (or cached) files
    return Bunch(
        signal=signal,
        left_steps=left_steps,
        right_steps=right_steps,
        metadata=metadata,
        description=DESCRIPTION,
    )
